/*
	Based upon a DDJ article (J.Spiller, Jul94.)
        Modifs by Benny.
	v0.3
*/
class MAT_COLUMN {
    float           Column[4];
 public:
    float          &operator[] (int Index) {
	return Column[Index];
    }
};

class MAT3D {
    MAT_COLUMN      Matrix[4];
 public:
    MAT_COLUMN & operator[] (int Index) {
	return Matrix[Index];
    }
    friend MAT3D operator *(MAT3D & A, MAT3D & B);
};

MAT3D operator *(MAT3D & A, MAT3D & B)
{
    int             I,    J,    K;
    MAT3D           Temp;

    for (I = 0; I < 4; I++)
	for (J = 0; J < 4; J++) {
	    float           Sum = 0;
	    for (K = 0; K < 4; K++)
		Sum += A[I][K] * B[K][J];
	    Temp[I][J] = Sum;
	}
    return Temp;
}

MAT3D           Identity()
{
    int             X,  Y;
    MAT3D           Temp;
    for (Y = 0; Y < 4; Y++)
        for (X = 0; X < 4; X++)
	    if (X == Y)
                Temp[Y][X] = 1;
	    else
                Temp[Y][X] = 0;
    return Temp;
}

MAT3D           RotateX(float Angle)
{
    MAT3D           Temp = Identity();
    Temp[1][1] = Temp[2][2] = cos(Angle);
    Temp[2][1] = -(Temp[1][2] = sin(Angle));
    return Temp;
}

MAT3D           RotateY(float Angle)
{
    MAT3D           Temp = Identity();
    Temp[0][0] = Temp[2][2] = cos(Angle);
    Temp[0][2] = -(Temp[2][0] = sin(Angle));
    return Temp;
}

MAT3D           RotateZ(float Angle)
{
    MAT3D           Temp = Identity();
    Temp[0][0] = Temp[1][1] = cos(Angle);
    Temp[1][0] = -(Temp[0][1] = sin(Angle));
    return Temp;
}

MAT3D           RotateXYZ(float AngleX, float   AngleY, float AngleZ)
{
    MAT3D       Temp3 = RotateX(AngleX),
		Temp2 = RotateY(AngleY),
		Temp1 = RotateZ(AngleZ);
    Temp2 = Temp2 * Temp3;
    return Temp1 * Temp2;
}

MAT3D           Scale(float Scale)
{
    MAT3D           Temp = Identity();
    Temp[0][0] = Temp[1][1] = Temp[2][2] = Scale;
    return Temp;
}

MAT3D           ScaleXYZ(float SX, float SY, float SZ)
{
    MAT3D           Temp = Identity();
    Temp[0][0] = SX;
    Temp[1][1] = SY;
    Temp[2][2] = SZ;
    return Temp;
}

MAT3D           Translate(float X, float Y, float Z)
{
    MAT3D           Temp = Identity();
    Temp[0][3] = X;
    Temp[1][3] = Y;
    Temp[2][3] = Z;
    return Temp;
}

void            Translate_Point(MAT3D &M, float &X, float &Y, float &Z)
{
    float TX,TY,TZ;
    TX = X * M[0][0] + Y * M[0][1] + Z * M[0][2] + M[0][3];
    TY = X * M[1][0] + Y * M[1][1] + Z * M[1][2] + M[1][3];
    TZ = X * M[2][0] + Y * M[2][1] + Z * M[2][2] + M[2][3];
    X = TX;    Y = TY;    Z = TZ;
}

int             IsVisible(MAT3D &M)
{
    //  Point1 = M *    [0, 0, 0, 1]t
    float           X1 = M[0][3];
    float           Y1 = M[1][3];
    float           Z1 = M[2][3];

    //  Point2 = M *    [0, 1, 0, 1]t
    float           X2 = M[0][1] + M[0][3];
    float           Y2 = M[1][1] + M[1][3];
    float           Z2 = M[2][1] + M[2][3];

    //  Point3 = M *    [1, 0, 0, 1]t
    float           X3 = M[0][0] + M[0][3];
    float           Y3 = M[1][0] + M[1][3];
    float           Z3 = M[2][0] + M[2][3];

    float    D =
    - X1 * (Y2 * Z3 - Y3 * Z2)
    - X2 * (Y3 * Z1 - Y1 * Z3)
    - X3 * (Y1 * Z2 - Y2 * Z1);

    if (!inside)
	return D > 0; // external view
    else
	return D < 0; // inside view
}
