/*
	2 rotating Flat/Iterated/Textured Cubes.
	Needs a (or two) .3df 256x256 textures in the same directory
	to work. Use the TexUS utility to convert any picture to .3df,
	available at www.3dfx.com.

        Glide 2.4, Watcom DOS.
        v.2
	March 99: updated for Banshee, because printf/puts
	messes up the Banshee graphics.

	To compile and link:
	 wcl386 [filename] -k512000 glide2x.lib
	To link only:
	 wlink f [filename] op stack=512000 n [filename] l clib3r,glide2x

	Bugs:
	- Cube will distort if too close to viewer (needs clipping)

-- Benny <benny_land@geocities.com>
*/
#include <dos.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h>
#include <stdio.h>
#include <conio.h>
#include <math.h>

#include <glide.h>

#define BANSHEE

#define YES		1
#define NO		0
#define ALL_FINE 	""

// Hardwired 3df x, y sizes (for now)
#define x_texsize		255
#define y_texsize		x_texsize

#define MAXFACE			6

GrHwConfiguration hwconfig;
short           GlideIsRunning = 0;

// startup preferences
short           inside = NO,
		iterated = YES,
		twosided = NO,
		allsides = YES,
		show3D = YES,
		decal = YES;

// Screen
long            size_x = 640, size_y = 480;

// Cube
float           side1 = 255,  side2 = 255;

int             red, gre, blu;
float           midx, midy;
int             grflatcolor;

long flatcolors[] ={
	 0xaa0101,
	 0xaa01aa,
	 0xaaaa01,
	 0x090909,
	 0x01aa01,
	 0x0101aa,
	 };

void quit(char *t);

#include "bentlib.h"		// modified tlib.c

FxU32 evenodd = GR_MIPMAPLEVELMASK_BOTH;
GrVertex        vtx1, vtx2, vtx3, vtx4;
GrFog_t         fogtable[GR_FOG_TABLE_SIZE];

Texture         map1Texture,map2Texture;
unsigned    	map1TextureAddr,map2TextureAddr;
Vertex3D      	Verts[4];

short 		which_texture = 0;

// --------------------------------------------------------------- //

#include "math3d.h"	// matrix functions
#include "usualgli.h"	// usual glide stuff and more

// --------------------------------------------------------------- //

// Change the vertexes colors around
void 		ChangeRGBs(int count)
{
    int j;
    for (j=0;j<count;j++) {
	red+=55;
	if (red>255) red -= 256;
	gre+=10;
	if (gre>255) gre -= 256;
	blu+=20;
	if (blu>255) blu -= 256;
	}
}

void            RenderFace(MAT3D &M, float Xp, float Yp)
{
    static lasttex = -1;
    //
    //  Project four corners
    //
    // upper left:              // (0,      0,      0) (in world space)
    vtx1.x = 0;
    vtx1.y = 0;
    vtx1.z = 0;
    // upper right:             // (height, 0,      0)
    vtx2.x = Xp;
    vtx2.y = 0;
    vtx2.z = 0;
    // lower left:              // (height, length, 0)
    vtx3.x = Xp;
    vtx3.y = Yp;
    vtx3.z = 0;
    // lower right:             // (0,      length, 0)
    vtx4.x = 0;
    vtx4.y = Yp;
    vtx4.z = 0;

    Translate_Point(M, vtx1.x, vtx1.y, vtx1.z);
    Translate_Point(M, vtx2.x, vtx2.y, vtx2.z);
    Translate_Point(M, vtx3.x, vtx3.y, vtx3.z);
    Translate_Point(M, vtx4.x, vtx4.y, vtx4.z);

    vtx1.oow = 1.f / vtx1.z;
    vtx2.oow = 1.f / vtx2.z;
    vtx3.oow = 1.f / vtx3.z;
    vtx4.oow = 1.f / vtx4.z;

    vtx1.x = vtx1.x * vtx1.oow + midx; // x1 (in screen space)
    vtx1.y = vtx1.y * vtx1.oow + midy; // y1
    vtx2.x = vtx2.x * vtx2.oow + midx; // x2
    vtx2.y = vtx2.y * vtx2.oow + midy; // y2
    vtx3.x = vtx3.x * vtx3.oow + midx; // x3
    vtx3.y = vtx3.y * vtx3.oow + midy; // y3
    vtx4.x = vtx4.x * vtx4.oow + midx; // x4
    vtx4.y = vtx4.y * vtx4.oow + midy; // y4

    SNAP(vtx1);SNAP(vtx2);SNAP(vtx3);SNAP(vtx4);

    //
    // Glide calls
    //
    if (!decal)
	if (iterated) {
		// iterated color...
		vtx1.r = red; vtx1.g = gre; vtx1.b = blu;    ChangeRGBs(1);
		vtx2.r = red; vtx2.g = gre; vtx2.b = blu;    ChangeRGBs(1);
		vtx3.r = red; vtx3.g = gre; vtx3.b = blu;    ChangeRGBs(1);
		vtx4.r = red; vtx4.g = gre; vtx4.b = blu;    ChangeRGBs(1);
	} else {
		// or flat color...
		grConstantColorValue(grflatcolor);
    } else {
	// or textured mapping:
	vtx1.tmuvtx[0].sow = Verts[0].s * x_texsize * vtx1.oow;
	vtx1.tmuvtx[0].tow = Verts[0].t * y_texsize * vtx1.oow;

	vtx2.tmuvtx[0].sow = Verts[1].s * x_texsize * vtx2.oow;
	vtx2.tmuvtx[0].tow = Verts[1].t * y_texsize * vtx2.oow;

	vtx3.tmuvtx[0].sow = Verts[2].s * x_texsize * vtx3.oow;
	vtx3.tmuvtx[0].tow = Verts[2].t * y_texsize * vtx3.oow;

	vtx4.tmuvtx[0].sow = Verts[3].s * x_texsize * vtx4.oow;
	vtx4.tmuvtx[0].tow = Verts[3].t * y_texsize * vtx4.oow;

	if (lasttex != which_texture)
	switch (which_texture) {
	case 0:
		grTexSource(GR_TMU0,
		    map1TextureAddr,
		    evenodd,
		    &map1Texture.info);
		break;
	case 1:
		grTexSource(GR_TMU0,
		    map2TextureAddr,
		    evenodd,
		    &map2Texture.info);
		break;
	}
	lasttex = which_texture;
    }

/*  Draw the triangles:
	1 2
	 /
	4 3     */

    guDrawTriangleWithClip(&vtx1, &vtx2, &vtx4);
    guDrawTriangleWithClip(&vtx2, &vtx4, &vtx3);
}

void StartGlideExtras()
{
    if (!decal)
	if (iterated)
	    // iterated colors
	    grColorCombine(GR_COMBINE_FUNCTION_LOCAL,
		GR_COMBINE_FACTOR_NONE,
		GR_COMBINE_LOCAL_ITERATED,
		GR_COMBINE_OTHER_NONE,
		FXFALSE);
	else
	    // flat shading
	    grColorCombine( GR_COMBINE_FUNCTION_LOCAL,
		 GR_COMBINE_FACTOR_NONE,
		 GR_COMBINE_LOCAL_CONSTANT,
		 GR_COMBINE_OTHER_NONE,
		 FXFALSE );
    else
	// textured
	grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER,
		   GR_COMBINE_FACTOR_ONE,
		   GR_COMBINE_LOCAL_NONE,
		   GR_COMBINE_OTHER_TEXTURE,
		   FXFALSE);

#define VX0	       0
#define VX1	       1

    // Let's use the full extend of the texture
    Verts[0].s = VX1; Verts[0].t = VX0;
    Verts[1].s = VX0; Verts[1].t = VX0;
    Verts[2].s = VX0; Verts[2].t = VX1;
    Verts[3].s = VX1; Verts[3].t = VX1;

    // Let's use the depth buffer
    grDepthBufferMode(GR_DEPTHBUFFER_WBUFFER);
    grDepthMask(FXTRUE); // needed (no default)

    // Let's use the mipmaps
    grTexMipMapMode(GR_TMU0,
		    GR_MIPMAP_NEAREST,
		    FXFALSE);

    // Let's go bilinear
    grTexFilterMode(GR_TMU0,
		    GR_TEXTUREFILTER_BILINEAR,
		    GR_TEXTUREFILTER_BILINEAR);

    /* No fog
    grAlphaBlendFunction(GR_BLEND_ONE, GR_BLEND_ZERO,
			 GR_BLEND_ONE, GR_BLEND_ZERO); */

    // Let's set the fog loose
#define FOGCOLOR 0xAA55FF //0xFFAA77
    grFogColorValue(FOGCOLOR);

    //.0001=far fog 	.01=close fog
    guFogGenerateExp(fogtable, .001);
    grFogTable(fogtable);
    grAlphaBlendFunction(GR_BLEND_ONE, GR_BLEND_PREFOG_COLOR,
			 GR_BLEND_ZERO, GR_BLEND_ZERO);
    grFogMode(GR_FOG_WITH_TABLE);

}

void info()
{
#ifndef BANSHEE
    char nfo[132];

    strcpy(nfo, "Mode: ");
    // View the inside/outside of the cube
    if (inside)	  strcat(nfo, "inside [v]iew, ");
    else       	  strcat(nfo, "outside [v]iew, ");
    // Draw invisible planes toggle
    if (twosided) strcat(nfo, "2 [s]ided, ");
    else          strcat(nfo, "1 [s]ided, ");
    //
    if (iterated) strcat(nfo, "iterated [c]olors, ");
    else          strcat(nfo, "flat [c]olors, ");
    // textured
    if (decal)	  strcat(nfo, "[t]extured, ");
    else      	  strcat(nfo, "no [t]exture, ");
    //
    if (allsides) strcat(nfo, "all [f]aces");
    else          strcat(nfo, "some [f]aces");

    printf("%s\n",nfo);
#endif
}

void            RenderAFace(int j, MAT3D &M)
{
    switch (j) {
    case 2:
    case 5:
	RenderFace(M, side2, side2);
	break;
    default:
	RenderFace(M, side1, side2);
	break;
    }
}

void	RenderCube(MAT3D Side[])
{
   int j;
   for (j=0;j<MAXFACE/2;j++) {
	grflatcolor = flatcolors[j];
	if (allsides && (twosided || IsVisible(Side[j])))
	    RenderAFace(j, Side[j]);
	else ChangeRGBs(4);
   }

   for (;j<MAXFACE;j++) {
	grflatcolor = flatcolors[j];
	if ((twosided || IsVisible(Side[j])))
	    RenderAFace(j, Side[j]);
	else ChangeRGBs(4);
   }
}

void mainloop()
{
    int      j = 0;
    int	     k, done = 0;
    MAT3D    S[MAXFACE];
    int	     running = 1;

    BuildCube(S);

#ifndef BANSHEE
    puts("\nCurrent options:");
    info();
#endif
    while (!done) {
	MAT3D   Cube1, Cube2, Temp;
	MAT3D 	Side1[MAXFACE], Side2[MAXFACE];
	if (running) {
	j++;

	Cube1 = Translate(-128, -100, -100);
	Temp = RotateXYZ(j * 0.01, j *  0.001, j * 0.007);
	Cube1 = Temp * Cube1;
	Temp = Translate(0, 0, sin(-j * 0.0121) * 350 + 600);
	Cube1 = Temp * Cube1;
	Temp = ScaleXYZ(200, 200, 1);
	Cube1 = Temp * Cube1;

	Cube2 = Translate(-300, -100, -100);
	Temp = RotateXYZ(j * 0.01, j *  0.001, j * 0.009);
	Cube2 = Temp * Cube2;
	Temp = Translate(0, 0, sin(-j * 0.00521) * 250 + 600);
	Cube2 = Temp * Cube2;
	Temp = ScaleXYZ(200, 200, 1);
	Cube2 = Temp * Cube2;

	for (k=0;k<MAXFACE;k++)
		Side1[k] = Cube1 * S[k];
	for (k=0;k<MAXFACE;k++)
		Side2[k] = Cube2 * S[k];
	}

	grBufferClear(FOGCOLOR, 0, GR_WDEPTHVALUE_FARTHEST);

	red = 200; gre = 0; blu = 150;
	which_texture = 0;

	midx = size_x/2 + size_x/16; // off-center
	RenderCube(Side1);

	which_texture = 1;

	midx = size_x/2 - size_x/16; // off-center
	RenderCube(Side2);

	grBufferSwap(1);

	if (kbhit()) switch (getch()) {
	    case 0x1b:		// VK_ESCAPE:
		done = YES;
		break;
	    case 0xD:		// VK_RETURN:
	    case 's':
		twosided = !twosided;
		info();
		break;
	    case 0x8:		// VK_BACK:
	    case 'v':
		inside = !inside;
		info();
		break;
	    case ' ':		// VK_SPACE:
		running = !running;
		break;
	    case 'c':
		iterated = !iterated;
		if (!iterated)
		    /* flat shading */
		    grColorCombine(GR_COMBINE_FUNCTION_LOCAL,
				   GR_COMBINE_FACTOR_NONE,
				   GR_COMBINE_LOCAL_CONSTANT,
				   GR_COMBINE_OTHER_NONE,
				   FXFALSE);
		else
		    /* iterated colors */
		    grColorCombine(GR_COMBINE_FUNCTION_LOCAL,
				   GR_COMBINE_FACTOR_NONE,
				   GR_COMBINE_LOCAL_ITERATED,
				   GR_COMBINE_OTHER_NONE,
				   FXFALSE);
		decal = NO;
		info();
		break;
	    case '+':
	    case 'f':
		allsides = !allsides;
		info();
		break;
#ifndef BANSHEE
	    // no go on a Banshee?
	    case 9:		// VK_TAB:
		show3D = !show3D;
		if (show3D)
		    grSstControl(GR_CONTROL_ACTIVATE);	 // 3dfx on
		else
		    grSstControl(GR_CONTROL_DEACTIVATE); // 3dfx off
		break;
#endif
	    case 't':
	    /* decal */
	    grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER,
		   GR_COMBINE_FACTOR_ONE,
		   GR_COMBINE_LOCAL_NONE,
		   GR_COMBINE_OTHER_TEXTURE,
		   FXFALSE);
	    decal = YES;
	    info();
	    break;
	    }
	if (done) break;
    }
}

void            main()
{
    puts("\n=== Multi-Texture Spinning Cubes ===");

#ifndef BANSHEE
    // If running on one monitor:
    char *ptr = getenv("SST_DUALHEAD");
    if (!ptr || ptr[0]=='0') {
	puts("\nOptions:\nPress TAB during viewing to see the options\n");
	delay(1000);
	}
#endif

    generalinit();
    mainloop();
    quit(ALL_FINE);
}
/* End */
