/*
    Usual Glide & Program Stuff.h
    (code that I change less often.)
    v0.2.1
*/

FxU32 textureSize, nextTexture, lastTexture;

void StartGlideExtras();

/*
	Snap functions
*/

void lowfpu()
{
long memvar;
_asm {
finit              // initialize the FPU
fwait              // wait for operation to complete
fstcw [memvar]     // store FPU control word to memvar
fwait              // wait for operation to complete
mov eax, [memvar]  // move memvar to a register
and eax, 0fffffcffh // mask off precision bits to set to 24-bit precision
mov [memvar], eax  // save control word to memory
fldcw [memvar]     // load control word back to FPU
fwait              // wait for operation to complete
}
}

#define SNAPPER (float)(3L << 18)
#define SNAP(vertex) \
(vertex).x += SNAPPER; (vertex).x -= SNAPPER; \
(vertex).y += SNAPPER; (vertex).y -= SNAPPER;

/*
   exit routine
*/
void quit(char *t)
{
	if (GlideIsRunning)
		grGlideShutdown();
	GlideIsRunning = 0;
	if (*t) { printf("\n=== Error: %s!\n",t); exit(1); }

	puts("\nBye.");
	exit(0);
}

/*
     Matrix cube.
     S1-S2-S3 are adjacent
*/
void BuildCube(MAT3D S[])
{
  MAT3D     Flip,Temp;
  int 	    j = 0;

    Flip = RotateY(180. * 3.141592654 / 180.);
    Temp = Translate(side1, 0, 0);
    Flip = Temp * Flip;

    S[j] = Translate(0, 0, 0);

    j++;
    Temp = RotateX(-90. * 3.141592654 / 180.);
    S[j] = Temp * Flip;

    j++;
    Temp = Translate(side2-side1, 0, 0);
    S[j] = Temp * Flip;
    Temp = RotateY(90. * 3.141592654 / 180.);
    S[j] = Temp * S[j];

    j++;
    Temp = Translate(0, 0, side2);
    S[j] = Temp * Flip;

    j++;
    S[j] = RotateX(-90. * 3.141592654 / 180.);
    Temp = Translate(0, side2, 0);
    S[j] = Temp * S[j];

    j++;
    S[j] = RotateY(90. * 3.141592654 / 180.);
    Temp = Translate(side1, 0, 0);
    S[j] = Temp * S[j];
}

/*
     Inits
*/

void StartGlide()
{
    grGlideInit();
    if (!(grSstQueryHardware(&hwconfig)))
	quit("Query Hardware failed");
    grSstSelect(0);
    if (!(grSstWinOpen(0,
			GR_RESOLUTION_640x480,
			GR_REFRESH_60Hz,
			GR_COLORFORMAT_ABGR,
			GR_ORIGIN_LOWER_LEFT, //faces outside
//			GR_ORIGIN_UPPER_LEFT, //faces inside
			2,
			1
			))) quit("Win Open failed");
    GlideIsRunning = YES;
}

void initbounds()
{
    nextTexture = grTexMinAddress(GR_TMU0);
    lastTexture = grTexMaxAddress(GR_TMU0);
}

/*
    Texture routines
*/

long getStartAddress(FxU32 evenOdd, GrTexInfo *info)
{
    long start;
    static int Totalsize = 0;
    static short initboundsflag = 0;

    if (!initboundsflag) {
	initbounds();
	initboundsflag = 1;
	}

    //sum of the sizes of the even/odd/all LODs
    textureSize = grTexTextureMemRequired(evenOdd, info);

    Totalsize += textureSize;
#ifndef BANSHEE
    printf("Total texture memory used = %i (%ik)\n",Totalsize,Totalsize/1024);
#endif

    start = nextTexture;
    nextTexture += textureSize;
    if (nextTexture <= lastTexture)
		return start;

    nextTexture = start;

    return -1;
}

// Based upon LoadTexture()
void DownLoad(char *tex, Texture *mapTexture, unsigned long &mapTextureAddr)
{
    if (!(LoadTexture(  tex,
			 &mapTexture->info,
			 &mapTexture->tableType,
			 &mapTexture->tableData)))
	quit("Load texture failed");
#ifndef BANSHEE
    printf("=== Read '%s' : ", tex);
    printf("width = %i, height = %i\n", info_width, info_height);
#endif
    if (info_width != x_texsize+1 || info_height != y_texsize+1) {
	char err[132];
	sprintf(err,"3df texture isn't %ix%i",x_texsize+1,y_texsize+1);
	quit(err);
	}

    mapTextureAddr = getStartAddress(evenodd, &mapTexture->info);

    if (mapTextureAddr == -1)
	quit("getStartAddress: out of texture mem "
		"(handling not implement yet!)");

    // Download texture data to TMU
    grTexDownloadMipMap(GR_TMU0,
			mapTextureAddr,
			evenodd,
			&mapTexture->info);
    if (map1Texture.tableType != NO_TABLE)
	grTexDownloadTable(GR_TMU0,
			   mapTexture->tableType,
			   &mapTexture->tableData);
}

void memcpytex(Texture *map1, Texture *map2)
{
    GrTexInfo      info;
    GrTexTable_t   tableType;
    TextureTable   tableData;

    memcpy(&map2->info,      &map1->info, sizeof(GrTexInfo));
    memcpy(&map2->tableType, &map1->tableType, sizeof(GrTexTable_t));
    memcpy(&map2->tableData, &map1->tableData, sizeof(TextureTable));
}

void ReadTex()
{
    struct find_t ffblk;
    unsigned done;

    done = _dos_findfirst("*.3df",_A_NORMAL,&ffblk);
    if (done) quit("no *.3df found");

    // TEX 1
    DownLoad(ffblk.name, &map1Texture, map1TextureAddr);

    // TEX 2
    done = _dos_findnext(&ffblk);
    if (done) {
	//quit("second *.3df not found");
	map1TextureAddr = map2TextureAddr;
	memcpytex(&map1Texture, &map2Texture);
	}
    else
	DownLoad(ffblk.name, &map2Texture, map2TextureAddr);

    // set up
    grTexCombine(GR_TMU0,
		     GR_COMBINE_FUNCTION_LOCAL,
		     GR_COMBINE_FACTOR_NONE,
		     GR_COMBINE_FUNCTION_LOCAL,
		     GR_COMBINE_FACTOR_NONE,
		     FXFALSE,
		     FXFALSE);
    grAlphaBlendFunction(GR_BLEND_ONE, GR_BLEND_ZERO,
			 GR_BLEND_ONE, GR_BLEND_ZERO);
}

/*
     Called by main()
*/
void generalinit()
{
    StartGlide();
    StartGlideExtras();
    lowfpu();

    midx = size_x/2;
    midy = size_y/2;

    ReadTex(); 		// read two textures
}
/*---*/
